/*******************************************************************************
**	IBM RGB526 PLL driver
********************************************************************************
**
   Copyright (C)	1998		Paul Sargent
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
********************************************************************************
**	This code was derived from the following sources of information:
**
**	[1]	Texas Intruments TVP 3026 PLL driver for KGI 
**		(C) 1997 Jonas Borgstrom & Steffen Seeger
**	[2]	IBM RGB526/RGB526DB Preliminary Datasheet 95/09/25 
**		(C) 1995 International Business Machines Corp.
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	"Paul Sargent <Paul.Sargent@3Dlabs.com>"
#define	DRIVER_NAME	"IBM 526 PLL driver"
#define	DRIVER_REV	"$Revision: 1.6 $"

#define	DEBUG_LEVEL	0

#include <kgi/module.h>
#include "pll.h"

#include "IBM/rgb52x.h"

static struct kgi_clock clock =
{
	"IBM",				/* manufacturer		*/
	"RGB526",			/* model		*/
	GM_ALL,				/* supported modes 	*/
	0,				/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/
	NULL,				/* Private data hook	*/
	CLK_TYPE_PROG,			/* clock type		*/
	{{
		8125000, 170000000,	/* clock range		*/
	}}
};

static const struct pll_limits dclk_pll = 
{
	14318180,			/* fref				*/
	{ 1, 1 },			/* factor		 	*/
	{ 65000000, 170000000 },	/* fvco range			*/
	{ 0, 3 },			/* post-VCO divider range	*/
	{ 3, 128 },			/* dividend range		*/
	{ 1, 14 }			/* divisor range		*/
};

#define	INVALID_COMMAND	((kgi_uint) -1)

static kgi_uint pll_command(struct pll_state *s, struct kgi_mode *mode)
{
	ASSERT(s->p   >= dclk_pll.p.min    &&  s->p   <= dclk_pll.mul.max);
	ASSERT(s->mul >= dclk_pll.mul.min  &&  s->mul <= dclk_pll.mul.max);
	ASSERT(s->div >= dclk_pll.div.min  &&  s->div <= dclk_pll.div.max);

	return (s->mul) | ((s->div) << 7) | (s->p << 13);
}

static void pll_set(struct kgi_display *dpy, kgi_uint clock_cmd)
{
   	kgi_uint pvalues[4] = {0, 1, 2, 4};
	kgi_uint pp, pm, pn, pc;

	/*	extract PCLK and LCLK settings from clock_cmd.
	*/
	pm = clock_cmd & 0x7F;
	pn = (clock_cmd >> 7) & 0x3F;
	pp = (clock_cmd >> 13) & 0x03;
 	
   	if ((14.31818 * pm / pn) > 65) {
	   
		pc = 3;
	} else {
	   
	   	pc = 2;
	}

	pp = pvalues[pp];       /* Translate to RGB stype M,N,P Values */
	pm--; pn--;
   
   	DEBUG1("Setting Pixel Clock m=%i,%.2x, n=%i,%.2x, p=%.2x, c=%.2x, cmd=%i",
		pm, pm, pn, pn, pp, pc, clock_cmd);
 	
   	/* Copy currently used values to Reg set 3 */
 	EXT_out8(dpy, EXT_in8(dpy, DAC_EXT_PIXEL_N_INPUT), DAC_EXT_PIXEL_N3);
 	EXT_out8(dpy, EXT_in8(dpy, DAC_EXT_PIXEL_M_INPUT), DAC_EXT_PIXEL_M3);
 	EXT_out8(dpy, EXT_in8(dpy, DAC_EXT_PIXEL_P_INPUT), DAC_EXT_PIXEL_P3);
 	EXT_out8(dpy, EXT_in8(dpy, DAC_EXT_PIXEL_C_INPUT), DAC_EXT_PIXEL_C3);

   	/* Select Standard Programming Scheme and Reg set 3 */
      	EXT_out8(dpy, _RGB52x_EXT10_PLL_STANDARD_REG, DAC_EXT_PIXEL_PLL_CTRL1);
 	EXT_out8(dpy, 0x03, DAC_EXT_PIXEL_PLL_CTRL2);

	/* Program new values into Reg set 2 */
	EXT_out8(dpy, pn, DAC_EXT_PIXEL_N2);
 	EXT_out8(dpy, pm, DAC_EXT_PIXEL_M2);
 	EXT_out8(dpy, pp, DAC_EXT_PIXEL_P2);
 	EXT_out8(dpy, pc, DAC_EXT_PIXEL_C2);

   	/* Select Standard Programming Scheme and Reg set 2 */
 	EXT_out8(dpy, _RGB52x_EXT10_PLL_STANDARD_REG, DAC_EXT_PIXEL_PLL_CTRL1);
 	EXT_out8(dpy, 0x02, DAC_EXT_PIXEL_PLL_CTRL2);

}

/*
**	KGI module interface
*/

#define	NO_KGIM_CLOCK_INIT
#define	NO_KGIM_CLOCK_DONE
#define	NO_KGIM_CLOCK_CHECK_MODE
#include "pll.inc"

kgi_sint kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	kgi_uint id,revision;
   
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	id       = EXT_in8(dpy, DAC_EXT_ID);
	revision = EXT_in8(dpy, DAC_EXT_REVISION);

	if (id == DAC_EXT01_RGB526_ID) {

		DEBUG0("Detected RGB526 RAMDAC. ID: 0x%.2x REV: 0x%.2x",id ,revision);
	
	} else {

		DEBUG0("unknown chip ID (%.2x).", id);
		return -E(RAMDAC, UNKNOWN);
	}

	card->clock = &clock;
	return EOK;
}

void kgim_clock_done(struct kgi_display *dpy)
{
	dpy->card->clock = NULL;
	NOTICE(DRIVER_NAME " removed");
}

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

kgi_sint kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
	enum kgi_tm_cmd *cmd)
{
	struct pll_state state;
	/*	We always have to provide DCLK for internal reference.
	*/
	kgi_uint newfreq = (*cmd == CMD_PROPOSE) ? 0 : mode->dclk;

	switch (*cmd) {

	case CMD_PROPOSE:
		mode->clock = INVALID_COMMAND;
		return EOK;

	case CMD_LOWER:

		if (newfreq < MIN_FREQ) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too low",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq > MAX_FREQ) {

			newfreq = MAX_FREQ;
		}

		if (pll_best_state(newfreq, *cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f;
		return EOK;

	case CMD_RAISE:

		if (MAX_FREQ < newfreq) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too high", \
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq < MIN_FREQ) {

			newfreq = MIN_FREQ;
		}

		if (pll_best_state(newfreq, *cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f;
		return EOK;

	case CMD_CHECK:

		mode->clock = INVALID_COMMAND;
		if (pll_best_state(newfreq, CMD_LOWER, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->clock = pll_command(&state, mode);
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
}
